#
# sims2_bodypainter.r
# Generates texture and alpha for the "bodypaint"-like meshes
# By: Alberto Vieira Ferreira Monteiro
# (aka albmont in many forums or mailing lists)
# Suggestions are Welcome!
#
# How to Use:
# (a) Install R, a free (both meaning: freedom and no cost) software
#     with powerful statistical tools and a very simple interface
# (b) Run R
# (c) Load this program with source("sims2_bodypainter.r")
# (d) Run the function bodypainter() passing arguments as desired
#     This step will create files texture0.png and alpha0.png
# (e) Start Bodyshop
# (f) Create a new project based on a bodypaint-like mesh
# (g) Edit the texture and the alpha bmp files, replacing them by 
#     texture0 and alpha0 from item (d)
# (h) Export the project to the game and play
#
# Step (d) will require more attention: some arguments are self-explanatory
# (and most of them aren't implemented), but in order to be complete,
# see below
#
# Modification History
# 2006-10-01 : initial version
# 2006-10-03 : panties as triangles or concave quadrilaterals
# 2006-10-05 : stars painted over the body
# 2006-10-06 : top cuts at side of body instead of going through arm
# 2006-10-07 : shirt
# 2006-10-10 : bug fixed: left and right were inverted
# 2006-10-11 : soutien.top
# 2006-10-12 : trousers;
#   border color (border.col) may be different from internal color (col)
# 2006-10-13 : improved match between graphic coordinates and R coordinates
#              (see below in par() for details)
# 2006-10-21 : shorts
# 2006-10-22 : fix border.col for shirt
#              
#

scale <- 1024 / 1450 # conversion from milimeters to pixels: 1450 mm ~ 1024 pixels

#
# bodypainter
# Probably this is the (only) function that must be called. Notice
# that R arguments to a function can be passed in order, or
# can be passed using the default name. For example, if sub(a, b)
# is the function that calculates a - b, you can call it by
# x <- sub(10, 3), x <- sub(a = 10, b = 3) or x <- sub(b = 3, a = 10).
# The (optional) arguments are:
#
# arg - input arguments. This is the return value
# user.begin and user.end - user provided functions. See below
# gender, age - gender and age. Right now, only gender = "f" and age = "a"
#   are meaningful
# ... - all other arguments (clothes to be added)
#
# example:
#
## Default panties:
#
# bodypainter(panties = 0)
#
## Default panties, and add a written text to them. Notice that, after the texture
## is drawn, the user.end function is called, and it "finishes" the picture.
## The arguments of myfunc must be arg, gender, age and type.
#
# myfunc <- function(arg, gender, age, type) { if (type == "t") { text(766, 620, "bodypainter") }; return(arg) }
# bodypainter(user.end = myfunc, panties = 0)
#
## Red panties (argument col, for color) with 10% transparency (argument alpha). 
## alpha = 1 is totally opaque
#
# bodypainter(panties = list(col = "red", alpha = 0.9))
#
# Each type of clothing will have more detailed information in the
# proper section. For example, draw.panties is the routine to draw
# panties.
#
# Current clothes:
# panties : bikini-bottom like
# top : a rectangle over the breast
# stars : stars all over the body [user must input positions]
# shirt : sleeveless shirt
#
bodypainter <- function(arg = NULL, user.begin = NULL, user.end = NULL,
    gender = NULL, age = NULL, ...)
{
  # create texture0.png
  arg <- bodypainter.file(arg, user.begin, user.end, gender, age, type = "t", ...)
  # create alpha0.png
  arg <- bodypainter.file(arg, user.begin, user.end, gender, age, type = "a", ...)
  # exit
  return(arg)
}

#
# Create files for texture and alpha, and draw them
#
bodypainter.file <- function(arg, user.begin, user.end, gender, age, type,
    panties = NULL, trousers = NULL, 
    top = NULL, soutien.top = NULL, shirt = NULL, stars = NULL, shorts = NULL)
{
  if (type == "t")
    png("texture0.png", width=1024, height=1024, bg="black")
  else if (type == "a")
    png("alpha0.png", width=1024, height=1024, bg="black")
  else
    return(arg)

# par and plot define coordinates from 0 to 1023 _inclusive_ (total of 1024 pixels)
  par(mar=c(0,0,0,0)) # reset margins
  plot(0, xlim=c(0, 1023), ylim=c(0, 1023), col="black")
  par(usr=c(0, 1023, 0, 1023))

# user-defined begin function
  if (!is.null(user.begin)) arg <- user.begin(arg, gender, age, type)

# do stuff here
  if (!is.null(trousers))
    arg <- draw.trousers(arg, gender, age, type, trousers)
  if (!is.null(shorts))
    arg <- draw.shorts(arg, gender, age, type, shorts)
  if (!is.null(panties))
    arg <- draw.panties(arg, gender, age, type, panties)
  if (!is.null(top))
    arg <- draw.top(arg, gender, age, type, top)
  if (!is.null(soutien.top))
    arg <- draw.soutien.top(arg, gender, age, type, soutien.top)
  if (!is.null(shirt))
    arg <- draw.shirt(arg, gender, age, type, shirt)
  if (!is.null(stars))
    arg <- draw.stars(arg, type, stars)

# user-defined end function
  if (!is.null(user.end)) arg <- user.end(arg, gender, age, type)

  dev.off()
  return(arg)
}    

#
# Draw panties
# panties is the description of the panties. It's a structure
# (created like: list(alpha = 0.75, col = "cyan")) with the description
# of the panties. The fields are:
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
# border.col : color of the border and the string parts. Defaults to col
#
# front.low : a value of 0 means that the waistline is horizontal.
#   positive values mean that the waistline gets lower by that ammount
#   of _milimeters_, scale based on an approximation. Defaults to 50
#
# front.angle : half-opening angle, in degrees, of the bottom of the panties. 
#   Defauts to 30
#
# front.shape : "t" for triangular, "q" for a (concave) quadrilateral,
#   "c" for a curved shape. Defaults to "t"
#
# back.low : same as front.low, for the back. Defaults to 20
# back.angle : same as front.angle, for the back. Defaults to 90
# back.shape : same as front.shape
#
draw.panties <- function(arg, gender, age, type, panties)
{
  if (type == "t") { # texture
    col <- ifelse(is.null(panties$col), "white", panties$col)
    border.col <- ifelse(is.null(panties$border.col), col, panties$border.col)
  }
  else {
    alpha <- ifelse(is.null(panties$alpha), 1.0, panties$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }

  y.comp <- get.panties.complement(gender, age)

  # front
  front.low <- ifelse(is.null(panties$front.low), 50.0, panties$front.low)
  front.angle <- pi / 180 * ifelse(is.null(panties$front.angle), 30.0, panties$front.angle)
  front.shape <- ifelse(is.null(panties$front.shape), "t", panties$front.shape)
  p1 <- get.front.panties.base(gender, age)
  p2 <- get.front.hip(gender, age, "r")
  p4 <- get.front.hip(gender, age, "l")
  p3 <- (p2 + p4) / 2 - c(0, front.low * scale)
  if (front.shape == "q") {
    px <- generate.shape.q(p1, p2, p3, p4, front.angle)
    xy <- cbind(p1, px$p2x, p3, px$p4x)
    polygon(xy[1,], xy[2,], col=col, border=NA)
    xy <- cbind(p2, p3, p4)
    points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  }
  else if (front.shape == "c") {
    xy0 <- cbind(p2, p3, p4)
    xy <- spline(xy0[1,], xy0[2,])
# ... continuar daqui ...
  }
  else {
    px <- generate.shape.t(p1, p3, front.angle)
    xy <- cbind(p1, px$p2x, px$p4x)
    polygon(xy[1,], xy[2,], col=col, border=NA)
    xy <- cbind(p2, px$p2x, px$p4x, p4)
    points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  }
  xy <- cbind(px$p2x, p1, px$p4x)
  points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  points(c(p1[1], p1[1]), c(p1[2], y.comp), col=border.col, type="l", lwd=2)

  # back
  back.low <- ifelse(is.null(panties$back.low), 20.0, panties$back.low)
  back.angle <- pi / 180 * ifelse(is.null(panties$back.angle), 45.0, panties$back.angle)
  back.shape <- ifelse(is.null(panties$back.shape), "t", panties$back.shape)
  p1 <- get.back.panties.base(gender, age)
  p2 <- get.back.hip(gender, age, "l")
  p4 <- get.back.hip(gender, age, "r")
  p3 <- (p2 + p4) / 2 - c(0, back.low * scale)
  if (back.shape == "q") {
    px <- generate.shape.q(p1, p2, p3, p4, back.angle)
    xy <- cbind(p1, px$p2x, p3, px$p4x)
    polygon(xy[1,], xy[2,], col=col, border=NA)
    xy <- cbind(p2, p3, p4)
    points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  }
  else {
    px <- generate.shape.t(p1, p3, back.angle)
    xy <- cbind(p1, px$p2x, px$p4x)
    polygon(xy[1,], xy[2,], col=col, border=NA)
    xy <- cbind(p2, px$p2x, px$p4x, p4)
    points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  }
  xy <- cbind(px$p2x, p1, px$p4x)
  points(xy[1,], xy[2,], col=border.col, type="l", lwd=2)
  points(c(p1[1], p1[1]), c(p1[2], y.comp), col=border.col, type="l", lwd=2)

  return(arg)
}

generate.shape.q <- function(p1, p2, p3, p4, angle) 
{
  # get coordinates of side of panties (ignoring lace)
  # (x1 - p1[1]) / (y1 - p1[2]) = tan(angle)
  # (x1 - p2[1]) / (y1 - p2[2]) = (p3[1] - p2[1]) / (p3[2] - p2[2]) = tan2
  # or:
  # x1 - tan(angle) y1 = p1[1] - p1[2] tan(angle) 
  # x1 - tan2 y1 = p2[1] - p2[2] tan2
  tan2 <- (p3[1] - p2[1]) / (p3[2] - p2[2])
  p2x <- solve(matrix(c(1, 1, -tan(angle), -tan2), 2, 2), 
    c(p1[1] - p1[2] * tan(angle), p2[1] - p2[2] * tan2))
  tan4 <- (p3[1] - p4[1]) / (p3[2] - p4[2])
  p4x <- solve(matrix(c(1, 1, tan(angle), -tan4), 2, 2), 
    c(p1[1] + p1[2] * tan(angle), p4[1] - p4[2] * tan4))
  return(list(p2x = p2x, p4x = p4x))
}

generate.shape.t <- function(p1, p3, angle)
{
  delta <- c(tan(angle) * (p3[2] - p1[2]), 0)
  return(list(p2x = p3 + delta, p4x = p3 - delta))
}

#
# Draw top
# top is the description of a rectangular (cylindrical when
# wrapped over the body) top. It's a structure
# (created like: list(alpha = 0.75, col = "cyan")) with the description
# of the top. The fields are:
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
# border.col : color of the border and the string parts. Defaults to col
#
# up, down : how many milimiters over and under the center of the
#   nipple it covers
#
draw.top <- function(arg, gender, age, type, top)
{
  if (type == "t") {
    col <- ifelse(is.null(top$col), "white", top$col)
    border.col <- ifelse(is.null(top$border.col), col, top$border.col)
  }
  else {
    alpha <- ifelse(is.null(top$alpha), 1.0, top$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }
  # y-coordinates
  mid <- 0.5 * (get.nipple(gender, age, "l") + get.nipple(gender, age, "r"))
  up <- ifelse(is.null(top$up), 20, top$up) * scale
  down <- ifelse(is.null(top$down), 20, top$down) * scale
  y <- c(mid[2] + up, mid[2] + up, mid[2] - down, mid[2] - down)
  # x-coordinates
  lbx <- get.torso.side(gender, age, side = "l", face = "b")
  rbx <- get.torso.side(gender, age, side = "r", face = "b")
  x.b <- c(lbx, rbx, rbx, lbx)
  polygon(x.b, y, col=col, border=NA)
  lfx <- get.torso.side(gender, age, side = "l", face = "f")
  rfx <- get.torso.side(gender, age, side = "r", face = "f")
  x.f <- c(lfx, rfx, rfx, lfx)
  polygon(x.f, y, col=col, border=NA)
  points(x.b[1:2], y[1:2], col=border.col, type="l", lwd=2)
  points(x.b[3:4], y[3:4], col=border.col, type="l", lwd=2)
  points(x.f[1:2], y[1:2], col=border.col, type="l", lwd=2)
  points(x.f[3:4], y[3:4], col=border.col, type="l", lwd=2)
  return(arg)
}

#
# draw.stars
# Paint stars at user-input positions
#
# Structure:
# x, y: center of the stars in pixels
# r: radius of the stars in pixels
# col: colors of the stars in pixels
# theta: orientation of the stars in pixels
#
draw.stars <- function(arg, type, stars) 
{
  x <- stars$x
  y <- stars$y
  if (is.null(x) | is.null(y)) return
  nx <- length(x)
  ny <- length(y)
  n <- min(nx, ny)
  if (nx > n) x <- x[1:n]
  if (ny > n) y <- y[1:n]
  r <- stars$r
  if (length(r) == 0) r <- 10
  if (length(r) < n) r <- rep(r, n)
  r <- r[1:n]
  theta <- stars$theta
  if (length(theta) == 0) theta <- 0
  if (length(theta) < n) theta <- rep(theta, n)
  theta <- theta[1:n]
  if (type == "a") 
    col <- rep("white", n)
  else { 
    col <- stars$col
    if (is.null(col)) col = "gray"
    if (length(col) < n) col <- rep(col, n)
    col <- col[1:n]
  }
  phi <- (1 + sqrt(5))/2
  kr <- rep(c(1, 1/phi^2), 5)
  j <- 1:10
  for (i in 1:n) {
    x1 <- x[i] + r[i] * kr * cos(theta[i] + j * pi / 5)
    y1 <- y[i] + r[i] * kr * sin(theta[i] + j * pi / 5)
    polygon(x1, y1, col = col[i], border=NA)
  }
  return(arg)
} 

#
# draw.shirt
# Draw a shirt, usually covering the nipples (all deltas are relative
# to their position, normally those deltas would be positive)
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
# border.col : color of the border and the string parts. Defaults to col
#
# front.raise and back.raise : raise the lower part of the arc, resp.
#   in the front or the back, by that number of milimeters. Default
#   is 0 for front and 100 for back.
#
# left.raise and right.raise : raise the side part (under the armpit)
#   by that number of milimeters. Default is 0.
#
# left.in, left.out, right.in, right.out : how much the part that goes
#   over the shoulder is thick. <side>.in is in the direction of the neck,
#   <side>.out is in the direction of the arm. Default is 0 milimeters
#   because there's a bug I can't trace
#
# cover.down : cover that number of milimeters under the line of the
#   nipples. Default is 100.
#
draw.shirt <- function(arg, gender, age, type, shirt)
{
  if (type == "t") {
    col <- ifelse(is.null(shirt$col), "white", shirt$col)
    border.col <- ifelse(is.null(shirt$border.col), col, shirt$border.col)
  }
  else {
    alpha <- ifelse(is.null(shirt$alpha), 1.0, shirt$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }
  front.raise <- ifelse(is.null(shirt$front.raise), 0, shirt$front.raise) * scale
  back.raise <- ifelse(is.null(shirt$back.raise), 100, shirt$back.raise) * scale
  l.raise <- ifelse(is.null(shirt$left.raise), 0, shirt$left.raise) * scale
  r.raise <- ifelse(is.null(shirt$right.raise), 0, shirt$right.raise) * scale
  cover.down <- ifelse(is.null(shirt$cover.down), 100, shirt$cover.down) * scale
  l.in <- ifelse(is.null(shirt$left.in), 0, shirt$left.in) * scale
  l.out <- ifelse(is.null(shirt$left.out), 0, shirt$left.out) * scale
  r.in <- ifelse(is.null(shirt$right.in), 0, shirt$right.in) * scale
  r.out <- ifelse(is.null(shirt$right.out), 0, shirt$right.out) * scale
  y.max <- get.max.y.torso(gender, age)
  for (face in c("f", "b")) {
    xy <- NULL
    border.xy <- NULL
    l.side.x <- get.torso.side(gender, age, side = "l", face = face)
    l.nip <- get.nipple(gender, age, side = "l")
    r.nip <- get.nipple(gender, age, side = "r")
    r.side.x <- get.torso.side(gender, age, side = "r", face = face)
    this.center <- get.panties.base(gender, age, face)
    this.raise <- ifelse(face == "f", front.raise, back.raise)
    if (face == "b") {
      f.center <- get.bellybutton(gender, age)
      l.nip[1] <- this.center[1] - (l.nip[1] - f.center[1])
      r.nip[1] <- this.center[1] - (r.nip[1] - f.center[1])
      # inversion of l/r in/out
      l.in <- -l.in
      l.out <- -l.out
      r.in <- -r.in
      r.out <- -r.out
    }
    xy1 <- ellipse.arc(c(l.side.x, y.max), 
      l.nip[1] - l.side.x + l.out, l.nip[2] - y.max + l.raise, origin = "y")
    xy <- cbind(xy1$x, xy1$y)
    border.xy <- rbind(xy, c(NA, NA))

    xy1 <- ellipse.arc(c(this.center[1], y.max),
      l.nip[1] - this.center[1] + l.in, l.nip[2] - y.max + this.raise, origin = "x")
    xy <- rbind(xy, cbind(xy1$x, xy1$y))
    border.xy <- rbind(border.xy, cbind(xy1$x, xy1$y), c(NA, NA))

    xy1 <- ellipse.arc(c(this.center[1], y.max),
      r.nip[1] - this.center[1] - r.in, r.nip[2] - y.max + this.raise, origin = "y")
    xy <- rbind(xy, cbind(xy1$x, xy1$y))
    border.xy <- rbind(border.xy, cbind(xy1$x, xy1$y), c(NA, NA))

    xy1 <- ellipse.arc(c(r.side.x, y.max), 
      r.nip[1] - r.side.x - r.out, r.nip[2] - y.max + r.raise, origin = "x")
    xy <- rbind(xy, cbind(xy1$x, xy1$y))
    border.xy <- rbind(border.xy, cbind(xy1$x, xy1$y), c(NA, NA))

    xy1 <- list(x = c(r.side.x, l.side.x), 
      y = c(r.nip[2] - cover.down, l.nip[2] - cover.down))
    xy <- rbind(xy, cbind(xy1$x, xy1$y))
    border.xy <- rbind(border.xy, cbind(xy1$x, xy1$y), c(NA, NA))

    polygon(xy, col=col, border=NA)
    points(border.xy, col=border.col, type="l", lwd=2)
  }
  return(arg)
}

#
# draw.soutien.top
# Draw a soutien-like top (bra-like top), usually covering the nipples
# (all deltas are relative to their position, normally those deltas
# would be positive). The core part of the soutien top is a triangle,
# the shoulder straps join the back forming right angles with the
# horizontal strap.
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
# border.col : color of the border and the string parts. Defaults to col
#
# upper.raise and upper.out : raise and move outwards the upper vertex of 
#   the triangle by that number of milimeters. Default is (75,0)
#
# outer.lower and outer.out : lower and move outwards the outer vertex of
#   the triangle by that number of milimeters. Default is (50, 50)
#
# inner.lower and inner.in : lower and move inwards the inner vertex of
#   the triangle by that number of milimeters. Default is (50, 50)
#
#
draw.soutien.top <- function(arg, gender, age, type, soutien.top)
{
  if (type == "t") {
    col <- ifelse(is.null(soutien.top$col), "white", soutien.top$col)
    border.col <- ifelse(is.null(soutien.top$border.col), col, soutien.top$border.col)
  }
  else {
    alpha <- ifelse(is.null(soutien.top$alpha), 1.0, soutien.top$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }
  upper.raise <- ifelse(is.null(soutien.top$upper.raise), 75, soutien.top$upper.raise) * scale
  upper.out <- ifelse(is.null(soutien.top$upper.out), 0, soutien.top$upper.out) * scale
  outer.lower <- ifelse(is.null(soutien.top$outer.lower), 50, soutien.top$outer.lower) * scale
  outer.out <- ifelse(is.null(soutien.top$outer.out), 50, soutien.top$outer.out) * scale
  inner.lower <- ifelse(is.null(soutien.top$inner.lower), 50, soutien.top$inner.lower) * scale
  inner.in <- ifelse(is.null(soutien.top$inner.in), 50, soutien.top$inner.in) * scale
  y.max <- get.max.y.torso(gender, age)
  for (face in c("f", "b")) {
    x <- NULL
    y <- NULL
    l.side.x <- get.torso.side(gender, age, side = "l", face = face)
    l.nip <- get.nipple(gender, age, side = "l")
    r.nip <- get.nipple(gender, age, side = "r")
    r.side.x <- get.torso.side(gender, age, side = "r", face = face)
    this.center <- get.panties.base(gender, age, face)
    if (face == "f") {
      # left side 
      l.upper <- l.nip + c(upper.out, upper.raise)
      l.inner <- l.nip + c(-inner.in, -inner.lower)
      l.outer <- l.nip + c(outer.out, -outer.lower)
      xy <- rbind(l.upper, l.inner, l.outer)
      polygon(xy, col=col, border=NA)
      # right side
      r.upper <- r.nip + c(-upper.out, upper.raise)
      r.inner <- r.nip + c(inner.in, -inner.lower)
      r.outer <- r.nip + c(-outer.out, -outer.lower)
      xy <- rbind(r.upper, r.inner, r.outer)
      polygon(xy, col=col, border=NA)
    }
    # lines
    l.side <- c(l.side.x, l.outer[2])
    r.side <- c(r.side.x, r.outer[2])
    l.shoulder <- c(l.upper[1], y.max)
    r.shoulder <- c(r.upper[1], y.max)
    na2 <- c(NA, NA)
    if (face == "f") {
      xy <- rbind(l.side, l.outer, l.upper, l.shoulder, na2, 
        l.outer, l.inner, l.upper, na2, l.inner, r.inner, na2, 
         r.side, r.outer, r.upper, r.shoulder, na2, r.outer, r.inner, r.upper)
      points(xy, col=border.col, type="l", lwd=2)
      arg$soutien.top$xy <- xy  # debug
    }
    else {   # face == "b": invert left/right
      f.center <- get.bellybutton(gender, age)
      l.tie <- c(this.center[1] - (l.upper[1] - f.center[1]), l.outer[2])
      r.tie <- c(this.center[1] - (r.upper[1] - f.center[1]), r.outer[2])
      l.side <- c(l.side.x, l.outer[2])
      r.side <- c(r.side.x, r.outer[2])
      l.shoulder <- c(l.tie[1], y.max)
      r.shoulder <- c(r.tie[1], y.max)
      xy <- rbind(l.side, r.side, na2, l.tie, l.shoulder, na2, r.tie, r.shoulder)
      points(xy, col=border.col, type="l", lwd=2)
      arg$soutien.top$xy.b <- xy  # debug
    }
  }
  return(arg)
}


#
# Draw trousers
# trousers is the description of the trousers. It's a structure
# (created like: list(alpha = 0.75, col = "cyan")) with the description
# of the trousers. The fields are:
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
draw.trousers <- function(arg, gender, age, type, trousers)
{
  if (type == "t") { # texture
    col <- ifelse(is.null(trousers$col), "white", trousers$col)
    border.col <- ifelse(is.null(trousers$border.col), col, trousers$border.col)
  }
  else {
    alpha <- ifelse(is.null(trousers$alpha), 1.0, trousers$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }

  for (face in c("f", "b")) {
    xy <- rbind(
      get.trousers.limit(age, gender, side = "l", face = face, up.down = "u"),
      get.trousers.limit(age, gender, side = "l", face = face, up.down = "d"),
      get.trousers.limit(age, gender, side = "lc", face = face, up.down = "d"),
      get.trousers.limit(age, gender, side = "c", face = face, up.down = 0),
      get.trousers.limit(age, gender, side = "rc", face = face, up.down = "d"),
      get.trousers.limit(age, gender, side = "r", face = face, up.down = "d"),
      get.trousers.limit(age, gender, side = "r", face = face, up.down = "u"))
    polygon(xy, col = col, border = NA)
  }
  return(arg)
}


#
# Draw shorts
# shorts is the description of the shorts. It's a structure
# (created like: list(alpha = 0.75, col = "cyan")) with the description
# of the shorts. The fields are:
#
# alpha : transparency factor. 1 = totally opaque, 0 = totally invisible
#
# col : color. Can be a string ("white", "pink", etc), or can be
#   created by the color-generating routines (rgb, hsv, etc).
#
# border.col : color of the border. Defaults to col
#
# front.low : a value of 0 means that the waistline is horizontal.
#   positive values mean that the waistline gets lower by that ammount
#   of _milimeters_, scale based on an approximation. Defaults to 50
#
# front.shape : "t" for triangular, "s" for sinusoidal. Defaults to "t"
#
# back.low : same as front.low, for the back. Defaults to 20
# back.shape : same as front.shape
#
# leg.down : distance, in milimeters, from the crotch down. Defaults to 20.
#
draw.shorts <- function(arg, gender, age, type, shorts)
{
  if (type == "t") { # texture
    col <- ifelse(is.null(shorts$col), "white", shorts$col)
    border.col <- ifelse(is.null(shorts$border.col), col, shorts$border.col)
  }
  else {
    alpha <- ifelse(is.null(shorts$alpha), 1.0, shorts$alpha)
    col <- hsv(0, 0, alpha)
    border.col <- "white"
  }

  low <- shape <- NULL  # to be defined below
  low[["f"]] <- ifelse(is.null(shorts$front.low), 50.0, shorts$front.low) * scale
  shape[["f"]] <- ifelse(is.null(shorts$front.shape), "t", shorts$front.shape)
  low[["b"]] <- ifelse(is.null(shorts$back.low), 20.0, shorts$back.low) * scale
  shape[["b"]] <- ifelse(is.null(shorts$back.shape), "t", shorts$back.shape)
  leg.down <- ifelse(is.null(shorts$leg.down), 20.0, shorts$leg.down) * scale
  hip <- get.front.hip(gender, age, "r")
  hip.y <- hip[2]

  for (face in c("f", "b")) {  # "front" and "back"
    tlu <- get.trousers.limit(age, gender, side = "l", face = face, up.down = "u")
    tld <- get.trousers.limit(age, gender, side = "l", face = face, up.down = "d")
    tlcd <- get.trousers.limit(age, gender, side = "lc", face = face, up.down = "d")
    tc <- get.trousers.limit(age, gender, side = "c", face = face, up.down = 0)
    trcd <- get.trousers.limit(age, gender, side = "rc", face = face, up.down = "d")
    trd <- get.trousers.limit(age, gender, side = "r", face = face, up.down = "d")
    tru <- get.trousers.limit(age, gender, side = "r", face = face, up.down = "u")

    plu <- intercept.segments(c(0, hip.y), c(1023, hip.y), tlu, tld)
    pru <- intercept.segments(c(0, hip.y), c(1023, hip.y), tru, trd)
    if (shape[[face]] == "t") {
      pc <- (plu + pru) / 2 - c(0, low[[face]])
      xy <- rbind(plu, pc, pru)
    }
    else {  # sinusoidal
      n <- abs(pru[1] - plu[1])
      x <- seq(plu[1], pru[1], length=n)
      y <- plu[2] + low[[face]] * (cos(seq(0, 2 * pi, length=n))-1) / 2
      xy <- cbind(x, y)
    }
    border.xy <- xy
    low.y <- tc[2] - leg.down
    prd <- intercept.segments(c(0, low.y), c(1023, low.y), tru, trd)
    prcd <- intercept.segments(c(0, low.y), c(1023, low.y), tc, trcd)
    xy <- rbind(xy, prd, prcd)
    border.xy <- rbind(border.xy, c(NA, NA), prd, prcd, c(NA, NA))
    plcd <- intercept.segments(c(0, low.y), c(1023, low.y), tc, tlcd)
    pld <- intercept.segments(c(0, low.y), c(1023, low.y), tlu, tld)
    xy <- rbind(xy, tc, plcd, pld)
    border.xy <- rbind(border.xy, plcd, pld, c(NA, NA))
    polygon(xy, col = col, border = NA)
    points(border.xy, col=border.col, type="l", lwd=2)
  }
  return(arg)
}




#
# Most graphic editors define (0,0) as the left _upper_ corner
# y = 1023 - y(i) to get R-like coordinates
#

#
# ellipse.arc returns the (x,y) coordinates of the 90 degrees
# arc from center + (delta.x, 0) to center + (0, delta.y)
#
ellipse.arc <- function(center, delta.x, delta.y, origin = "x")
{
  if (is.null(origin)) origin <- "x"
  theta <- seq(0, pi/2, length = min(abs(delta.x), abs(delta.y)))
  if (origin == "y") theta <- pi/2 - theta
  return(list(x = center[1] + delta.x * cos(theta),
    y = center[2] + delta.y * sin(theta)))
}

ellipse_points <- function(center, ab, theta)
{
  return(list(x = center[1] + ab[1] * cos(theta),
    y = center[2] + ab[2] * sin(theta)))
}

#
# Retrieve coordinates from the standard image of the naked body
#

get.panties.base <- function(gender, age, face)
{
  if (face == "f") return(get.front.panties.base(gender, age))
  if (face == "b") return(get.back.panties.base(gender, age))
  return(c(0,0))
}

get.front.panties.base <- function(gender, age)
{
  return(c(768, 1023 - 471))  # gender = f, age = a/t
}

get.front.hip <- function(gender, age, side)
{
  return(c(ifelse(side == "r", 868, 664), 1023 - 369))  # gender = f, age = a
}

get.back.panties.base <- function(gender, age)
{
  return(c(256, 1023 - 404))
}

get.back.hip <- function(gender, age, side)
{
  return(c(ifelse(side == "r", 152, 363), 1023 - 369))
}



get.panties.complement <- function(gender, age)
{
  return(1023 - 481) # gender = f, age = a
}

get.nipple <- function(gender, age, side)
{
  return(c(ifelse(side == "r", 716, 820), 1023 - 175))
}

# (734,401)

get.torso.side <- function(gender, age, side, face)
{
  f.hip <- get.front.hip
  if (face == "b") f.hip <- get.back.hip
  hip.this.x <- (f.hip(gender, age, side))[1]
  hip.other.x <- (f.hip(gender, age, ifelse(side == "l", "r", "l")))[1]
  torso.size <- ifelse(face == "f", 877 - 656, 367 - 142)
  middle <- (hip.this.x + hip.other.x) / 2
  side <- middle + sign(hip.other.x - hip.this.x) * torso.size / 2
  return(side)
}

get.max.y.torso <- function(gender, age)
{
  return(1023 - 35)
}

get.bellybutton <- function(gender, age)
{
  return(c(768, 1023 - 334))
}

get.trousers.limit <- function(gender, age, side, face, up.down)
{
  if (side == "c") {  # center
    return(if(face == "f") c(768, 1023 - 471) else c(254, 1023 - 471))
  }
  if (up.down == "u")
    y <- ifelse(face == "f", 1023 - 318, 1023 - 310)
  else if (side == "lc" | side == "rc")
    y <- ifelse(face == "f", 1023 - 870, 1023 - 863)
  else # (side == "l" | side == "r") & (up.down == "d")
    y <- ifelse(face == "f", 1023 - 865, 1023 - 859)
  #
  # x
  #
  if (up.down == "u" && face == "b" && side == "r")
    x <- 148
  else if (up.down == "u" && face == "b" && side == "l")
    x <- 357
  else if (up.down == "d" && face == "b" && side == "r")
    x <- 139
  else if (up.down == "d" && face == "b" && side == "rc")
    x <- 250
  else if (up.down == "d" && face == "b" && side == "lc")
    x <- 257
  else if (up.down == "d" && face == "b" && side == "l")
    x <- 366
  else if (up.down == "u" && face == "f" && side == "l")
    x <- 663
  else if (up.down == "u" && face == "f" && side == "r")
    x <- 872
  else if (up.down == "d" && face == "f" && side == "l")
    x <- 654
  else if (up.down == "d" && face == "f" && side == "lc")
    x <- 765
  else if (up.down == "d" && face == "f" && side == "rc")
    x <- 773
  else if (up.down == "d" && face == "f" && side == "r")
    x <- 880
  #
  # and this only for FA!
  #
  return(c(x, y))
}

#
# intercept.segments interceps two segments. segments = TRUE mean that
# they are really segments, so solutions outside the segments are invalid
# segments = FALSE intercepts them as lines. No check is done for parallelism
# of the lines.
#
intercept.segments <- function(line1.p1, line1.p2, line2.p1, line2.p2, segments = TRUE) 
{
  # solve equations
  # line1.p1 + u1 (line1.p2 - line1.p1) = line2.p1 + u2 (line2.p2 - line2.p1)
  # subject to the restriction that u1 and u2 are between 0 and 1
  M <- cbind(line1.p2 - line1.p1, line2.p1 - line2.p2)
  b <- line2.p1 - line1.p1
  u <- solve(M, b)
  if (segments & (any(u < 0) | any(u > 1))) return(c(NA, NA))
  return(line1.p1 + u[1] * (line1.p2 - line1.p1))
}
